#!/usr/bin/python3
from __future__ import print_function
import os
# import IgorModel

import requests
from bs4 import BeautifulSoup
from Bio import SeqIO

try:
    from StringIO import StringIO # Python 2
except ImportError:
    from io import StringIO # Python 3

imgt_params={
    'url.home' : "http://www.imgt.org",
    'url.genelist' : "http://www.imgt.org/download/GENE-DB/IMGTGENEDB-GeneList",
    'url.genedb' : "http://www.imgt.org/genedb/GENElect?",
    'url.release' : "http://www.imgt.org/download/GENE-DB/RELEASE",
    'url.readme' : "http://www.imgt.org/download/GENE-DB/README.txt",
    'Genesymbol' : [],
    'data.chainlist' : ['TRA', 'TRB', 'BCR']
}

def get_species_list():
    """
    Returns list of all species available in the IMGT database.
    :return : species list to make queries to IMGT GENEDB.
    """
    response = requests.get(imgt_params['url.genelist'])
    tmp_species_list = set()
    for linea in response.text.split('\n')[1:]:
        tmp_species_list.add(linea.split(';')[0])
    species_list = [specie.replace(" ", "+") for specie in tmp_species_list]
    return species_list

def get_genedb_query72(specie: str, gene: str, imgt_genedb=imgt_params['url.genedb']):
    """
    Returns imgt link to download genomic template according to imgt database.
    :param specie: imgt specie name check get_species_list() to get the names of species in IMGT DB.
    :param gene: imgt nomenclature for gene.
    :return : string of the requested link
    """
    strQuery = "query=7.2+"
    url_gene = imgt_genedb + strQuery + gene + "&species=" + specie
    return url_gene

def get_genedb_query81_imgtlabel(specie: str, gene: str, imgtlabel:str, imgt_genedb=imgt_params['url.genedb']):
    """
    Returns imgt link to download genomic template according to imgt database with the corresponding IMGTlabel
    :return : string of the requested link
    """
    strQuery = "query=8.1+"
    url_gene = imgt_genedb + strQuery + gene + "&species=" + specie+"&IMGTlabel="+imgtlabel
    return url_gene

def extract_fasta_from_url(url):
    """ Extract fasta information from the typical IMGT format
    """
    print(url)
    # Get content from imgt url
    response = requests.get(url)

    soup = BeautifulSoup(response.text, 'html.parser')
    result = soup.find_all('pre')[1].text
    return result

#def genSeqRecords(url):
def get_records_list(url):
    """
    :return : a list of Sequence records.
    """
    rawfasta = extract_fasta_from_url(url)
    # FIXME: Is better to preserve the generator but ... you know maybe later XD
    return list(SeqIO.parse(StringIO(rawfasta), "fasta"))

def save_records2fasta(records, filename:str):
    # flnGenomicJs = DIR_REF_GENOME + "genomicJs.fasta"
    with open(filename, "w") as ofile:
        SeqIO.write(records, ofile, "fasta")

    return filename

def load_records_from_fasta(filename:str):
    # flnGenomicJs = DIR_REF_GENOME + "genomicJs.fasta"
    return list(SeqIO.parse(filename, "fasta"))

def makeDirectories(gene: str, specie: str, modelspath=None):
    if modelspath is None:
        modelspath = "models"

    os.system("mkdir -p " + modelspath)
    os.system("mkdir -p " + modelspath + "/" + specie)
    os.system("mkdir -p " + modelspath + "/" + specie + "/" + gene)
    os.system("mkdir -p " + modelspath + "/" + specie + "/" + gene + "/ref_genome")
    os.system("mkdir -p " + modelspath + "/" + specie + "/" + gene + "/ref_genome")
    os.system("mkdir -p " + modelspath + "/" + specie + "/" + gene + "/models")

def get_gene_template(specie: str, gene: str, modelspath=None, filename=None, imgt_genedb=imgt_params['url.genedb']):
    """
    Download genomic template according to specie and gene from imgt database
    :return : string of the requested link
    """
    try:
        url_query = get_genedb_query72(specie, gene, imgt_genedb=imgt_genedb)
        print(specie, gene, url_query)

        # 2. load data from IMGT url
        records = get_records_list(url_query)
        records = map(lambda x: x.upper(), records)
        return records
    except Exception as e:
        raise e

def download_gene_template(specie: str, gene: str, modelspath=None, filename=None, imgt_genedb=imgt_params['url.genedb']):
    """
    Create a file in IGoR default format
    and returns output filename.
    :param specie: IMGT specie nomenclature use "+" instead of a space " " Mus+musculus
    :param gene: IMGT gene nomenclature like TRAV, IGHJ, etc
    """
    # 3. Write it in a fasta file.
    records = get_gene_template(specie, gene, imgt_genedb=imgt_genedb)

    if gene[-1] in ['V', 'D', 'J']:
        ref_genes_path = ""
        if modelspath is None:
            modelspath = "models"
            gene_dir = gene[:-1]
            makeDirectories(gene_dir, specie, modelspath=modelspath)
            ref_genes_path = modelspath + "/" + specie + "/" + gene_dir + "/ref_genome/"

        if filename is None :
            filename = ref_genes_path + "genomic" + gene[-1] + "s__imgt.fasta"
            #filename = modelspath + "genomic__" + specie + "_" + gene + ".fasta"
    #flnGenomicJs = DIR_REF_GENOME + "genomicJs.fasta"
    flnGenome = save_records2fasta(records, filename)
    fln_dict = dict()
    fln_dict["fln"+gene[-1]+"Genome"] = flnGenome
    fln_dict['ref_genes_path'] = ref_genes_path
    fln_dict['modelspath'] = modelspath
    return fln_dict

def genKey(seqDescription):
    """
    :param seqDescritption : is the header description in fasta file
    :return key : a generated key to create a dictionary
    :return startPos : the start position of the sequence based on ACCesion number sequence.
    """
    fields = seqDescription.split("|")
    key    = "|".join(fields[:4])
    return key

def getStartPos(seqDescription):
    """
    Return the start position of sequence based on the fasta file description
    """
    fields = seqDescription.split("|")
    #startPosLIGMDB = map(int, fields[5].split(".."))[0]
    startPosLIGMDB = int ( fields[5].split("..")[0] )
    return startPosLIGMDB

def getFunction(seqDescription):
    """
    Return Functionality
    """
    fields = seqDescription.split("|")
    strFunction = fields[3]
    return strFunction

def genAnchDict(url):
    recordsAnch = get_records_list(url)
    seqsDictAnch = {}
    for rec in recordsAnch:
        seqsDictAnch[genKey(rec.description)] = getStartPos(rec.description)
    #print("genAnchDict", seqsDictAnch)
    return seqsDictAnch

def get_gene_anchors(specie: str, gene: str, imgtlabel: str, modelspath=None, filename=None, imgt_genedb=imgt_params['url.genedb']):
    """
    Download genomic anchors for V or J according to specie and gene from imgt database
    :return : string of the requested link
    """
    try:
        #QUERY_V_2CYS = "query=8.1+" + gene + "&species=" + specie + "&IMGTlabel=2nd-CYS"
        url_query = get_genedb_query81_imgtlabel(specie, gene, imgtlabel, imgt_genedb=imgt_genedb)
        print(specie, gene, url_query)

        # 2. load data from IMGT url
        records = get_records_list(url_query)
        records = map(lambda x: x.upper(), records)
        return records
    except Exception as e:
        raise e

def download_genes_anchors(specie: str, chain: str, flnVGenome, flnJGenome, modelspath=None, imgt_genedb=imgt_params['url.genedb']):
    # For this method anchors will be in the same directory modelspath+"/"+specie+"/"+chain+
    V_dict = download_Vgene_anchors(specie, chain, flnVGenome, modelspath=modelspath, imgt_genedb=imgt_genedb)
    J_dict = download_Jgene_anchors(specie, chain, flnJGenome, modelspath=modelspath, imgt_genedb=imgt_genedb)
    Anchors_dict = {**V_dict, **J_dict}
    return Anchors_dict

def download_Vgene_anchors(specie: str, chain: str, flnVGenome, modelspath=None, imgt_genedb=imgt_params['url.genedb']):
    #records = get_gene_template(specie, gene, imgt_genedb=imgt_genedb)
    if modelspath is None:
        modelspath = "models"
    makeDirectories(chain, specie, modelspath=modelspath)
    ref_genes_path = modelspath + "/" + specie + "/" + chain + "/ref_genome/"

    # generate dictionaries with the anchors
    # 2nd-CYS
    urlV_2CYS = get_genedb_query81_imgtlabel(specie, chain + "V", imgtlabel="2nd-CYS", imgt_genedb=imgt_genedb)
    dictV_2CYS = genAnchDict(urlV_2CYS)

    flnAnchors = ref_genes_path +  "V_gene_CDR3_anchors__imgt.csv"
    ofileAnch = open(flnAnchors, "w")
    ofileAnch.write("gene;anchor_index;function" + "\n")
    CSVDELIM = ";"

    Vrecords = SeqIO.parse(flnVGenome, "fasta")
    for rec in Vrecords:
        key = genKey(rec.description)
        # Initiate to avoid bad designation.
        posV_2CYS = -1
        if key in dictV_2CYS.keys():
            posV_2CYS = dictV_2CYS[key] - getStartPos(rec.description)
            posAnch = posV_2CYS
            seqFunc = getFunction(rec.description)
            ofileAnch.write(rec.description + CSVDELIM + str(posAnch) +  CSVDELIM+str(seqFunc)+"\n")
        else:
            print("No anchor is found for : " + rec.description)

    ofileAnch.close()
    fln_dict = dict()
    fln_dict['flnVGenome'] = flnVGenome
    fln_dict['flnVAnchors'] = flnAnchors
    fln_dict['ref_genes_path'] = ref_genes_path
    fln_dict['modelspath'] = modelspath
    return fln_dict

def download_Jgene_anchors(specie: str, chain: str, flnJGenome, modelspath=None, imgt_genedb=imgt_params['url.genedb']):
    # records = get_gene_template(specie, gene, imgt_genedb=imgt_genedb)
    if modelspath is None:
        modelspath = "models"

    makeDirectories(chain, specie, modelspath=modelspath)
    ref_genes_path = modelspath + "/" + specie + "/" + chain + "/ref_genome/"

    # J-PHE
    urlJ_PHE = get_genedb_query81_imgtlabel(specie, chain + "J", imgtlabel="J-PHE", imgt_genedb=imgt_genedb)
    dictJ_PHE = genAnchDict(urlJ_PHE)

    # J-TRP
    urlJ_TRP = get_genedb_query81_imgtlabel(specie, chain + "J", imgtlabel="J-TRP", imgt_genedb=imgt_genedb)
    dictJ_TRP = genAnchDict(urlJ_TRP)

    # write the files
    flnAnchors = ref_genes_path + "J_gene_CDR3_anchors__imgt.csv"
    ofileAnch = open(flnAnchors, "w")
    # ofileAnch.write("gene;anchor_index" + "\n")
    ofileAnch.write("gene;anchor_index;function" + "\n")
    CSVDELIM = ";"
    Jrecords = SeqIO.parse(flnJGenome, "fasta")
    for rec in Jrecords:
        key = genKey(rec.description)
        # Initiate to avoid bad designation.
        posJ_PHE = -1
        posJ_TRP = -1
        if key in dictJ_PHE.keys():
            posJ_PHE = dictJ_PHE[key] - getStartPos(rec.description)
            posAnch = posJ_PHE
            seqFunc = getFunction(rec.description)
            ofileAnch.write(rec.description + CSVDELIM + str(posAnch) + CSVDELIM + str(seqFunc) + "\n")
            # ofileAnch.write(rec.description + CSVDELIM + str(posAnch) + "\n")

        # FIXME: this should be elif, but I want to check If the system is consistent
        elif key in dictJ_TRP.keys():
            posJ_TRP = dictJ_TRP[key] - getStartPos(rec.description)
            posAnch = posJ_TRP
            seqFunc = getFunction(rec.description)
            ofileAnch.write(rec.description + CSVDELIM + str(posAnch) + CSVDELIM + str(seqFunc) + "\n")
            # ofileAnch.write(rec.description + CSVDELIM + str(posAnch) + "\n")
        else:
            print("No anchor is found for : " + rec.description)

    ofileAnch.close()
    fln_dict = dict()
    fln_dict['flnJGenome'] = flnJGenome
    fln_dict['flnJAnchors'] = flnAnchors
    fln_dict['ref_genes_path'] = ref_genes_path
    fln_dict['modelspath'] = modelspath
    return fln_dict

############ Donwload all genetic information if VDJ or VJ
def download_ref_genome_VDJ(species: str, chain: str, **kwargs):
    dictVGenome = download_gene_template(species, chain + 'V', **kwargs)
    dictDGenome = download_gene_template(species, chain + 'D', **kwargs)
    dictJGenome = download_gene_template(species, chain + 'J', **kwargs)

    flnVGenome = dictVGenome["flnVGenome"]
    flnDGenome = dictDGenome["flnDGenome"]
    flnJGenome = dictJGenome["flnJGenome"]

    # FIXME: ONCE THE GENE TEMPLATES ARE DOWNLOADED CHANGE THE NAME TO
    # write anchors
    Anchors_dict = download_genes_anchors(species, chain, flnVGenome, flnJGenome, **kwargs)

    # TODO: filter sequences for OLGA compatibility
    strGene = chain + "V"
    urlV_2CYS = get_genedb_query81_imgtlabel(species, strGene, imgtlabel="2nd-CYS")
    dict2CYS = genAnchDict(urlV_2CYS)
    list2CYS = list(dict2CYS.keys())
    # print("------------> list2CYS : ", list2CYS)

    v_genomes_list = load_records_from_fasta(flnVGenome)
    # print("------------D v_genomes_list: ", v_genomes_list)
    v_genomes_trim_list = list()
    for v_genome in v_genomes_list:
        if genKey(v_genome.description) in list2CYS:
        # if v_genome in list2CYS:
            v_genomes_trim_list.append(v_genome)
        else:
            print(v_genome.description)
    save_records2fasta(v_genomes_trim_list, flnVGenome + "_trim")

    # J-PHE
    urlJ_PHE = get_genedb_query81_imgtlabel(species, species + "J", imgtlabel="J-PHE")
    dictJ_PHE = genAnchDict(urlJ_PHE)
    listJ_PHE = list(dictJ_PHE.keys())

    # J-TRP
    urlJ_TRP = get_genedb_query81_imgtlabel(species, species + "J", imgtlabel="J-TRP")
    dictJ_TRP = genAnchDict(urlJ_TRP)
    listJ_TRP = list(dictJ_TRP.keys())

    j_genomes_list = load_records_from_fasta(flnJGenome)
    j_genomes_trim_list = list()
    for j_genome in j_genomes_list:
        if genKey(j_genome.description) in listJ_PHE:
        #if j_genome in listJ_PHE:
            j_genomes_trim_list.append(j_genome)
        elif genKey(j_genome.description) in listJ_TRP:
        # elif j_genome in listJ_TRP:
            j_genomes_trim_list.append(j_genome)
        else:
            print(j_genome.description)

    save_records2fasta(j_genomes_trim_list, flnJGenome + "_trim")
    print("----------------------")
    print("Genomic VDJ templates in files: ")
    print(flnVGenome, flnDGenome, flnJGenome)
    dict_V = gen_short_names(flnVGenome, flnAnchors=Anchors_dict["flnVAnchors"])
    dict_J = gen_short_names(flnJGenome, flnAnchors=Anchors_dict["flnJAnchors"])
    dict_D = gen_short_names(flnDGenome)

    import os.path
    V_path = os.path.dirname(dict_V["genomics"])
    J_path = os.path.dirname(dict_J["genomics"])
    D_path = os.path.dirname(dict_D["genomics"])
    import shutil
    shutil.copy(dict_V["genomics"], V_path + "/genomicVs.fasta")
    shutil.copy(dict_J["genomics"], J_path + "/genomicJs.fasta")
    shutil.copy(dict_D["genomics"], D_path + "/genomicDs.fasta")
    shutil.copy(dict_V["anchors"], V_path + "/V_gene_CDR3_anchors.csv")
    shutil.copy(dict_J["anchors"], J_path + "/J_gene_CDR3_anchors.csv")
    # print(kwargs)

def download_ref_genome_VJ(species: str, chain: str, **kwargs):
    dictVGenome = download_gene_template(species, chain + 'V', **kwargs)
    dictJGenome = download_gene_template(species, chain + 'J', **kwargs)

    flnVGenome = dictVGenome["flnVGenome"]
    flnJGenome = dictJGenome["flnJGenome"]

    # FIXME: ONCE THE GENE TEMPLATES ARE DOWNLOADED CHANGE THE NAME TO
    # write anchors
    Anchors_dict = download_genes_anchors(species, chain, flnVGenome, flnJGenome, **kwargs)

    # TODO: filter sequences for OLGA compatibility
    strGene = chain + "V"
    urlV_2CYS = get_genedb_query81_imgtlabel(species, strGene, imgtlabel="2nd-CYS")
    dict2CYS = genAnchDict(urlV_2CYS)
    list2CYS = list(dict2CYS.keys())
    # print("------------> list2CYS : ", list2CYS)

    v_genomes_list = load_records_from_fasta(flnVGenome)
    # print("------------D v_genomes_list: ", v_genomes_list)
    v_genomes_trim_list = list()
    for v_genome in v_genomes_list:
        if genKey(v_genome.description) in list2CYS:
        # if v_genome in list2CYS:
            v_genomes_trim_list.append(v_genome)
        else:
            print(v_genome)
    save_records2fasta(v_genomes_trim_list, flnVGenome + "_trim")

    # J-PHE
    urlJ_PHE = get_genedb_query81_imgtlabel(species, species + "J", imgtlabel="J-PHE")
    dictJ_PHE = genAnchDict(urlJ_PHE)
    listJ_PHE = list(dictJ_PHE.keys())

    # J-TRP
    urlJ_TRP = get_genedb_query81_imgtlabel(species, species + "J", imgtlabel="J-TRP")
    dictJ_TRP = genAnchDict(urlJ_TRP)
    listJ_TRP = list(dictJ_TRP.keys())

    j_genomes_list = load_records_from_fasta(flnJGenome)
    j_genomes_trim_list = list()
    for j_genome in j_genomes_list:
        if genKey(j_genome.description) in listJ_PHE:
        #if j_genome in listJ_PHE:
            j_genomes_trim_list.append(j_genome)
        elif genKey(j_genome.description) in listJ_TRP:
        # elif j_genome in listJ_TRP:
            j_genomes_trim_list.append(j_genome)
        else:
            print(j_genome)

    save_records2fasta(j_genomes_trim_list, flnJGenome + "_trim")

    print("----------------------")
    print("Genomic VJ templates in files: ")
    print(flnVGenome, flnJGenome)
    dict_V = gen_short_names(flnVGenome, flnAnchors=Anchors_dict["flnVAnchors"])
    dict_J = gen_short_names(flnJGenome, flnAnchors=Anchors_dict["flnJAnchors"])

    import os.path
    V_path = os.path.dirname(dict_V["genomics"])
    J_path = os.path.dirname(dict_J["genomics"])
    import shutil
    shutil.copy(dict_V["genomics"], V_path + "/genomicVs.fasta")
    shutil.copy(dict_J["genomics"], J_path + "/genomicJs.fasta")
    shutil.copy(dict_V["anchors"], V_path + "/V_gene_CDR3_anchors.csv")
    shutil.copy(dict_J["anchors"], J_path + "/J_gene_CDR3_anchors.csv")


def make_VDJ_model():
    from .IgorIO import IgorModel_Parms
    mdl_parms = IgorModel_Parms

def gen_short_names(flnGenome, flnAnchors=None):
    import pandas as pd
    from .IgorIO import genLabel
    # flnGenome = options.gene_template
    # flnAnchors = options.gene_anchors

    records = load_records_from_fasta(flnGenome)
    new_records = list()
    for record in records:
        record.description = genLabel(record.description)
        record.id = record.description
        new_records.append(record)
    flnGenome_short = flnGenome + "_short"
    save_records2fasta(new_records, flnGenome_short)

    gene_dict = dict()
    gene_dict["genomics"] = flnGenome_short

    try:
        if flnAnchors is not None:
            df = pd.read_csv(flnAnchors, sep=';')
            #print(df['gene'])
            df['gene'] = df['gene'].map(genLabel)
            # print(df['gene'])
            flnAnchors_short = flnAnchors + "_short"
            df.to_csv(flnAnchors_short, sep=';', index=False)
            gene_dict["anchors"] = flnAnchors_short
    except Exception as e:
        print(e)


    return gene_dict




# FIXME: PUT THIS IN __init__.py ? is a good practice?
# option wrappers for IGoR
# IGoR command line options
igor_option_path_dict={
    "alpha": "tcr_alpha",
    "beta" : "tcr_beta",
    "light": "light",
    "heavy_naive" : "bcr_heavy",
    "heavy_memory": "bcr_heavy"
}

# Options used to align in  IGoR
tmp_dict_options = {
    '---thresh': {'active': False, 'value': '15', 'dict_options': {}},
    '---matrix': {'active': False, 'value': 'path/to/file', 'dict_options': {}},
    '---gap_penalty': {'active': False, 'value': 'X', 'dict_options': {}},
    '---best_align_only': {'active': False, 'value': '', 'dict_options': {}}
}
igor_align_dict_options = {
    '--all':
        {'active': True, 'value': '',
         'dict_options':
             tmp_dict_options
         },
    '--V':
        {'active': False, 'value': '',
         'dict_options':
             tmp_dict_options
         },
    '--D':
        {'active': False, 'value': '',
         'dict_options':
             tmp_dict_options
         },
    '--J':
        {'active': False, 'value': '',
         'dict_options':
             tmp_dict_options
         }
}
igor_align_dict_options['--V']['dict_options']['---thresh']['value'] = '50'


igor_evaluate_dict_options = {
    '--N_iter':
        {'active': False, 'value': '5'},
    '--L_thresh':
        {'active': False, 'value': ''},
    '--P_ratio_thresh':
        {'active': False, 'value': '0.0'}, # Check all possible scenarios
    '--MLSO':
        {'active': False, 'value': ''},
    '--infer_only':
        {'active': False, 'value': ''},
    '--not_infer':
        {'active': False, 'value': ''},
    '--fix_err':
        {'active': False, 'value': ''},

}
# --N_iter N
# --L_thresh X
# --P_ratio_thresh X
# --MLSO
# --infer_only eventnickname1 eventnickname2
# --not_infer eventnickname1
# --fix_err

# Options used to output in  IGoR
igor_output_dict_options = {
    '--scenarios':
        {'active': True, 'value': '10',
         'dict_options':
             {}
         },
    '--Pgen':
        {'active': True, 'value': '',
         'dict_options':
             {}
         },
    '--coverage':
        {'active': False, 'value': '',
         'dict_options':
             {}
         }
}

# FIXME: MAKE A DICTIONARY
igor_file_id_list = [
# aligns directory
'indexed_sequences',
'indexed_CDR3',

'aligns_V_alignments',
'aligns_D_alignments',
'aligns_J_alignments',

# inference directory
'infer_final_parms',
'infer_final_marginals',

# evaluate directory
'evaluate_final_parms',
'evaluate_final_marginals',

# output directory
'output_pgen',
'output_scenarios',
'output_coverage',

# IGoR ref_genome
'genomicVs',
'genomicDs',
'genomicJs',

'V_gene_CDR3_anchors',
'J_gene_CDR3_anchors',

# IGoR models
'model_parms',
'model_marginals',

]

# dictionary for the status of the filenames used in IGoR at different stages.
igor_batch_dict = {
    # aligns directory
    'indexed_sequences' :  {'filename': "", 'status': False},
    'indexed_CDR3' :  {'filename': "", 'status': False},

    'aligns_V_alignments' :  {'filename': "", 'status': False},
    'aligns_D_alignments' :  {'filename': "", 'status': False},
    'aligns_J_alignments' :  {'filename': "", 'status': False},

    # inference directory
    'infer_final_parms' :  {'filename': "", 'status': False},
    'infer_final_marginals' :  {'filename': "", 'status': False},

    # evaluate directory
    'evaluate_final_parms' :  {'filename': "", 'status': False},
    'evaluate_final_marginals' :  {'filename': "", 'status': False},

    # output directory
    'output_pgen' :  {'filename': "", 'status': False},
    'output_scenarios' :  {'filename': "", 'status': False},
    'output_coverage' :  {'filename': "", 'status': False},

    # IGoR ref_genome
    'genomicVs' :  {'filename': "", 'status': False},
    'genomicDs' :  {'filename': "", 'status': False},
    'genomicJs' :  {'filename': "", 'status': False},

    'V_gene_CDR3_anchors' :  {'filename': "", 'status': False},
    'J_gene_CDR3_anchors' :  {'filename': "", 'status': False},

    # IGoR models
    'model_parms' :  {'filename': "", 'status': False},
    'model_marginals' :  {'filename': "", 'status': False}

}

def update_igor_batch_dict(igor_wd, igor_batchname):
    tmp_prefix_aligns = igor_wd + "/aligns/" + igor_batchname
    igor_batch_dict['indexed_sequences']['filename'] = tmp_prefix_aligns + "_indexed_sequences.csv"
    igor_batch_dict['indexed_CDR3']['filename'] = tmp_prefix_aligns + "_indexed_CDR3.csv"
    igor_batch_dict['aligns_V_alignments']['filename'] = tmp_prefix_aligns + "_V_alignments.csv"
    igor_batch_dict['aligns_D_alignments']['filename'] = tmp_prefix_aligns + "_D_alignments.csv"
    igor_batch_dict['aligns_J_alignments']['filename'] = tmp_prefix_aligns + "_J_alignments.csv"

    tmp_prefix = igor_wd + "/" + igor_batchname
    igor_batch_dict['infer_final_parms']['filename'] = tmp_prefix + "_inference/" + "final_parms.txt"
    igor_batch_dict['infer_final_marginals']['filename'] = tmp_prefix + "_inference/" + "final_marginals.txt"
    igor_batch_dict['evaluate_final_parms']['filename'] = tmp_prefix + "_evaluate/" + "final_parms.txt"
    igor_batch_dict['evaluate_final_marginals']['filename'] = tmp_prefix + "_evaluate/" + "final_marginals.txt"
    igor_batch_dict['output_pgen']['filename'] = tmp_prefix + "_output/" + "Pgen_counts.csv"
    igor_batch_dict['output_scenarios']['filename'] = tmp_prefix + "_output/" + "best_scenarios_counts.csv"
    igor_batch_dict['output_coverage']['filename'] = tmp_prefix + "_output/" + "coverage.csv"






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 20 12:24:34 2019

@author: alfaceor
"""
import pandas as pd

class IgorIndexedSequence:
    def __init__(self):
        self.seq_index     = -1
        self.sequence      = ""
        
    def to_dict(self):
        dictIndexedSequence       =  {
            "seq_index"     : self.seq_index  , \
            "sequence"      : self.sequence
            }
        
        return dictIndexedSequence
    
    @classmethod
    def load_FromCSVline(cls, csvline, delimiter=";"):
        cls = IgorIndexedSequence()
        csvsplit = csvline.replace("\n", "").split(";")
        try:
            cls.seq_index    = int  (csvsplit[0])
            cls.sequence     = int  (csvsplit[1])
        except Exception as e:
            print(e)
            raise e
        return cls
    
    @classmethod
    def load_FromSQLRecord(cls, sqlRecord):
        cls = IgorIndexedSequence()
        try:
            cls.seq_index    = int  (sqlRecord[0])
            cls.sequence     = int  (sqlRecord[1])
        except Exception as e:
            print(e)
            raise e
        return cls





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='pygor3',
    url="https://github.com/alfaceor/pygor3",
    author="Carlos Olivares",
    author_email="carlos.olivares@phys.ens.fr",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version='0.0.1',
    description='Python package to manipulate and run IGoR data files',
    license="GNU GPLv3",
    python_requires='>=3.5',
    install_requires = [
        "pandas",
        #"holoviz",
        "numpy",
        "xarray",
        "beautifulsoup4",
        "biopython",
        "Click",
        "networkx",
        "requests",
        "matplotlib",
        "airr"
    ],
    packages=setuptools.find_packages(),
    entry_points= {
        'console_scripts' : [
            'pygor=pygor3.scripts.cli:cli',
            'pygor3-bs_pairwise_prob=pygor3.scripts.bs_pairwise_prob:main',
            'pygor3-observable=pygor3.scripts.observable:main',
            'pygor3-scriptTest=pygor3.scripts.scriptTest:main',
            'pygor3-scriptTest02=pygor3.scripts.scriptTest02:main',
            'pygor3-stas_from_bs=pygor3.scripts.stas_from_bs:main',
            'pygor3-load_database=pygor3.scripts.load_database:main',
            'pygor3-naive_align=pygor3.scripts.naive_align:main',
            'pygor3-all_gene_alignments_id=pygor3.scripts.all_gene_alignments_id:main',
            'pygor3-naive_align_id=pygor3.scripts.naive_align_id:main',
            'pygor3-align_export=pygor3.scripts.align_export:main',
            'pygor3-bs_export=pygor3.scripts.bs_export:main',
            #'igor-pgen_sequences=pygor3.scripts.pgen_sequences:main',
            #'igor-infer_sequences=pygor3.scripts.infer_sequences:main',
            #'igor-model_plot=pygor3.scripts.model_plot:main',
            #'igor-model_export=pygor3.scripts.model_export:main',
            'pygor3-model_export=pygor3.scripts.model_export:main',
            'pygor3-plot_marginals=pygor3.scripts.plot_marginals:main',
            #'igor-model_create=pygor3.scripts.model_create:main',
            #'igor-get_imgt_data=pygor3.scripts.get_imgt_data:main',
            'pygor3-download_imgt_gene_templates=pygor3.scripts.download_imgt_gene_templates:main',
            'pygor3-short_names_imgt_templates=pygor3.scripts.short_names_imgt_templates:main'
            #'igor-model_import=pygor3.model_import:main',
            #'igor-get_imgt_data=pygor3.get_imgt_data:main'
        ],
    }
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  7 10:18:45 2019

@author: alfaceor
"""

from .IgorIO import *

class IgorBestScenariosVDJ:
    def __init__(self):
        self.seq_index = -1
        self.scenario_rank = -1
        self.scenario_proba_cond_seq = -1
        self.id_v_choice = -1
        self.id_j_choice = -1
        self.id_d_gene = -1
        self.id_v_3_del = -1
        self.id_d_5_del = -1
        self.id_d_3_del = -1
        self.id_j_5_del = -1
        self.id_vd_ins = -1
        self.vd_dinucl = list()
        self.id_dj_ins = -1
        self.dj_dinucl = list()
        self.mismatches = list()
        self.mismatcheslen = -1
        
        # Case of use of the IgorModel.Model_Parms class.
        self.flnModelParms = ""
        self.mdlParms = ""
        self.mdl = ""
        #self.IgorModelParms
        self.strSeq_index = ""
        
#        # To fetch data from datase connection to a database
#        self.IgorDB = ""

    def __str__(self):
        return str(self.to_dict())
    
    def setModel_Parms(self, flnModelParms):
        self.flnModelParms = flnModelParms
        self.mdlParms   = IgorModel_Parms(model_parms_file = self.flnModelParms)

    def to_dict(self):
        dictBestScenario       =  {
            "seq_index"     : self.seq_index     , \
            "scenario_rank" : self.scenario_rank , \
            "scenario_proba_cond_seq" : self.scenario_proba_cond_seq , \
            "v_choice"      : self.id_v_choice   , \
            "j_choice"      : self.id_j_choice   , \
            "d_gene"        : self.id_d_gene     , \
            "v_3_del"       : self.id_v_3_del    , \
            "d_5_del"       : self.id_d_5_del    , \
            "d_3_del"       : self.id_d_3_del    , \
            "j_5_del"       : self.id_j_5_del    , \
            "vd_ins"        : self.id_vd_ins     , \
            "vd_dinucl"     : self.vd_dinucl     , \
            "dj_ins"        : self.id_dj_ins     , \
            "dj_dinucl"     : self.dj_dinucl     , \
            "mismatches"    : self.mismatches    , \
            "mismatcheslen" : self.mismatcheslen
            }
        
        return dictBestScenario

    def to_dict_names(self):
        dictBestScenario = {
            "seq_index": self.seq_index, \
            "scenario_rank": self.scenario_rank, \
            "scenario_proba_cond_seq": self.scenario_proba_cond_seq, \
            "v_choice": self.getV_gene_name(), \
            "j_choice": self.getJ_gene_name(), \
            "d_gene": self.getD_gene_name(), \
            "v_3_del": self.getV_3_dels(), \
            "d_5_del": self.getD_5_dels(), \
            "d_3_del": self.getD_3_dels(), \
            "j_5_del": self.getJ_5_dels(), \
            "vd_ins": self.getVD_ins(), \
            "vd_dinucl": self.vd_dinucl, \
            "dj_ins": self.getDJ_ins(), \
            "dj_dinucl": self.dj_dinucl, \
            "mismatches": self.mismatches, \
            "mismatcheslen": self.mismatcheslen
        }

        return dictBestScenario

    def to_dict_values(self):
        dictBestScenario = {
            "seq_index": self.seq_index, \
            "scenario_rank": self.scenario_rank, \
            "scenario_proba_cond_seq": self.scenario_proba_cond_seq, \
            "v_choice": self.getV_gene_name(), \
            "j_choice": self.getJ_gene_name(), \
            "d_gene": self.getD_gene_name(), \
            "v_3_del": self.getV_3_dels(), \
            "d_5_del": self.getD_5_dels(), \
            "d_3_del": self.getD_3_dels(), \
            "j_5_del": self.getJ_5_dels(), \
            "vd_ins": self.getVD_ins(), \
            "vd_dinucl": self.getVD_Region(), \
            "dj_ins": self.getDJ_ins(), \
            "dj_dinucl": self.dj_dinucl, \
            "mismatches": self.mismatches, \
            "mismatcheslen": self.mismatcheslen
        }

        return dictBestScenario

    def to_dict_ntsequences(self):
        dictBestScenario = {
            "seq_index": self.seq_index, \
            "scenario_rank": self.scenario_rank, \
            "scenario_proba_cond_seq": self.scenario_proba_cond_seq, \
            "v_choice": self.getV_ntsequence(), \
            "j_choice": self.getJ_ntsequence(), \
            "d_gene": self.getD_ntsequence(), \
            "v_3_del": self.getV_3_dels(), \
            "d_5_del": self.getD_5_dels(), \
            "d_3_del": self.getD_3_dels(), \
            "j_5_del": self.getJ_5_dels(), \
            "vd_ins": self.getVD_ins(), \
            "vd_dinucl": self.getVD_Region(), \
            "dj_ins": self.getDJ_ins(), \
            "dj_dinucl": self.getDJ_Region(), \
            "mismatches": self.mismatches, \
            "mismatcheslen": self.mismatcheslen
        }

        return dictBestScenario

    # FIXME: CHANGE FOR A BETTER NAME
    def to_dict_export(self):
        dictBestScenario = {
            "seq_index": self.seq_index, \
            "scenario_rank": self.scenario_rank, \
            "scenario_proba_cond_seq": self.scenario_proba_cond_seq, \
            "v_choice": self.getV_ntsequence(), \
            "j_choice": self.getJ_ntsequence(), \
            "d_gene": self.getD_ntsequence(), \
            "v_3_del": self.getV_3_dels(), \
            "d_5_del": self.getD_5_dels(), \
            "d_3_del": self.getD_3_dels(), \
            "j_5_del": self.getJ_5_dels(), \
            "vd_ins": self.getVD_ins(), \
            "vd_dinucl": self.getVD_Region(), \
            "dj_ins": self.getDJ_ins(), \
            "dj_dinucl": self.getDJ_Region(), \
            "mismatches": self.mismatches, \
            "mismatcheslen": self.mismatcheslen
        }

        return dictBestScenario


    @classmethod
    def load_FromLineBestScenario(cls, line, delimiter=";"):
        #seq_index;scenario_rank;scenario_proba_cond_seq;GeneChoice_V_gene_Undefined_side_prio7_size35;GeneChoice_J_gene_Undefined_side_prio7_size14;GeneChoice_D_gene_Undefined_side_prio6_size2;Deletion_V_gene_Three_prime_prio5_size21;Deletion_D_gene_Five_prime_prio5_size21;Deletion_D_gene_Three_prime_prio5_size21;Deletion_J_gene_Five_prime_prio5_size23;Insertion_VD_genes_Undefined_side_prio4_size31;DinucMarkov_VD_genes_Undefined_side_prio3_size16;Insertion_DJ_gene_Undefined_side_prio2_size31;DinucMarkov_DJ_gene_Undefined_side_prio1_size16;Mismatches
        cls = IgorBestScenariosVDJ()
        linesplit = line.split(delimiter)
        linesplit = line.split(";")
        for ii in range(len(linesplit)):
            # TODO: find a better way to do this, if is a list keep it as list
            if (ii in [ 11, 13, 14 ]):
                linesplit[ii] = linesplit[ii]
            else:
                linesplit[ii] = linesplit[ii].replace("(", "").replace(")", "")

        try:
            #1;1;0.123596;(14);(9);(1);(4);(8);(7);(11);(0);();(9);(0,2,0,1,2,3,2,0,0);(122,123,124)
            cls.seq_index      = int(linesplit[ 0])
            cls.scenario_rank  = int(linesplit[ 1])
            cls.scenario_proba_cond_seq = float(linesplit[2])
            #print(linesplit[ 3], type(linesplit[ 3]), len(linesplit[ 3]))
            cls.id_v_choice    = int(linesplit[ 3])
            cls.id_j_choice    = int(linesplit[ 4])
            cls.id_d_gene      = int(linesplit[ 5])
            cls.id_v_3_del     = int(linesplit[ 6])
            cls.id_d_5_del     = int(linesplit[ 7])
            cls.id_d_3_del     = int(linesplit[ 8])
            cls.id_j_5_del     = int(linesplit[ 9])
            cls.id_vd_ins      = int(linesplit[10])
            cls.vd_dinucl      = eval(linesplit[11].replace("(","[").replace(")","]"))
            cls.id_dj_ins      = int(linesplit[12])
            cls.dj_dinucl      = eval(linesplit[13].replace("(","[").replace(")","]"))
            cls.mismatches     = eval(linesplit[14].replace("(","[").replace(")","]"))
            cls.mismatcheslen  = int(len(cls.mismatches))
            
            return cls
        
        except Exception as e:
            print(e)
            raise e
    
    
    @classmethod
    def load_FromDict(cls, dictBestScenarios):
        """
        Return a IgorBestScenariosVDJ instance from a IgorSqlRecord.
        :param sqlRecordAlign: record of a sql database table.
        :param strGene_name: gene_name associated to the record.
        :return: IgorAlignment_data instance
        """
        cls = IgorBestScenariosVDJ()
        try:            
#            cls.seq_index      = int(sqlRecordBestScenarios[ 0])
#            cls.scenario_rank  = int(sqlRecordBestScenarios[ 1])
#            cls.scenario_proba_cond_seq = float(sqlRecordBestScenarios[2])
            cls.id_v_choice    = int(dictBestScenarios["v_choice"])
            cls.id_j_choice    = int(dictBestScenarios["j_choice"])
            cls.id_d_gene      = int(dictBestScenarios["d_gene"])
            cls.id_v_3_del     = int(dictBestScenarios["v_3_del"])
            cls.id_d_5_del     = int(dictBestScenarios["d_5_del"])
            cls.id_d_3_del     = int(dictBestScenarios["d_3_del"])
            cls.id_j_5_del     = int(dictBestScenarios["j_5_del"])
            cls.id_vd_ins      = int(dictBestScenarios["vd_ins"])
            cls.vd_dinucl      = eval(dictBestScenarios["vd_dinucl"])
            cls.id_dj_ins      = int(dictBestScenarios["dj_ins"])
            cls.dj_dinucl      = eval(dictBestScenarios["dj_dinucl"])
#            cls.mismatches     = eval(dictBestScenarios["mismatches"])
#            cls.mismatcheslen  = int(len(cls.mismatches) )

            return cls
        except Exception as e:
            print(e)
            raise e
    
    @classmethod
    def load_FromSQLRecord(cls, sqlRecordBestScenarios):
        """
        Return a IgorBestScenariosVDJ instance from a IgorSqlRecord.
        :param sqlRecordAlign: record of a sql database table.
        :param strGene_name: gene_name associated to the record.
        :return: IgorAlignment_data instance
        """
        cls = IgorBestScenariosVDJ()
        try:
            cls.seq_index      = int(sqlRecordBestScenarios[ 0])
            cls.scenario_rank  = int(sqlRecordBestScenarios[ 1])
            cls.scenario_proba_cond_seq = float(sqlRecordBestScenarios[2])
            cls.id_v_choice    = int(sqlRecordBestScenarios[ 3])
            cls.id_j_choice    = int(sqlRecordBestScenarios[ 4])
            cls.id_d_gene      = int(sqlRecordBestScenarios[ 5])
            cls.id_v_3_del     = int(sqlRecordBestScenarios[ 6])
            cls.id_d_5_del     = int(sqlRecordBestScenarios[ 7])
            cls.id_d_3_del     = int(sqlRecordBestScenarios[ 8])
            cls.id_j_5_del     = int(sqlRecordBestScenarios[ 9])
            cls.id_vd_ins      = int(sqlRecordBestScenarios[10])
            cls.vd_dinucl      = eval(sqlRecordBestScenarios[11])
            cls.id_dj_ins      = int(sqlRecordBestScenarios[12])
            cls.dj_dinucl      = eval(sqlRecordBestScenarios[13])
            cls.mismatches     = eval(sqlRecordBestScenarios[14])
            cls.mismatcheslen  = int(sqlRecordBestScenarios[15])

            return cls
        except Exception as e:
            print(e)
            raise e
    
    # TODO: finish this class
    @classmethod
    def load_FromEventNameValues(cls, mdl, seq_index, strSeq_index, scenario_dict):
        # v_3_del = 2 
        # d_5_del = 6
        # d_3_del = 1
        # vd_ins  = 1 and should be a "C"
        # dj_ins  = 3 and should be a "TCT"
        # FIXME: CORRECT ERROR MESSAGES and ADD documentation and VALIDATIONS OF EVENTS
        """
        Return a IgorBestScenariosVDJ instance from a dict of names or values.
        :param strGene_name: gene_name associated to the record.
        :return: IgorAlignment_data instance
        """
        ##### The event I think is the best one
        cls = IgorBestScenariosVDJ() #.load_FromSQLRecord(record_bs[ii])
        #cls.setModel_Parms(flnModelParms)
        cls.mdl = mdl
        cls.seq_index = seq_index #59
        cls.strSeq_index  = strSeq_index #db.fetch_IgorIndexedSeq_By_seq_index(seq_index)[1]
        Event_GeneChoice = ['v_choice', 'j_choice', 'd_gene']
        Event_Deletions  = ['v_3_del', 'd_5_del', 'd_3_del', 'j_5_del']
        Event_Insertions = ['vd_ins', 'dj_ins']
        Event_Dinucl     = ['vd_dinucl', 'dj_dinucl']
        for event_nickname in scenario_dict.keys():
            if event_nickname in Event_GeneChoice:
                pd_event  = cls.mdl.parms.Event_dict[event_nickname]
                gene_name =   scenario_dict[event_nickname] # 'TRBV17*01'
                gene_id   = pd_event.loc[pd_event['name'] == gene_name ].index.values[0]
                if event_nickname == 'v_choice':
                    cls.id_v_choice = gene_id
                elif event_nickname == 'j_choice':
                    cls.id_j_choice = gene_id
                elif event_nickname == 'd_gene':
                    cls.id_d_gene = gene_id
                else:
                    print( "Something vey bad happen with "+str(scenario_dict[event_nickname]) )
                
            elif event_nickname in Event_Deletions:
                pd_event    = cls.mdl.parms.Event_dict[event_nickname]
                realiz_name = scenario_dict[event_nickname]
                realiz_id   = pd_event.loc[pd_event['value'] == realiz_name ].index.values[0]
                if event_nickname == 'v_3_del':
                    cls.id_v_3_del = realiz_id
                elif event_nickname == 'd_5_del':
                    cls.id_d_5_del = realiz_id
                elif event_nickname == 'd_3_del':
                    cls.id_d_3_del = realiz_id
                elif event_nickname == 'j_5_del':
                    cls.id_j_5_del = realiz_id
                else:
                    print( "Something vey bad happen with "+str(scenario_dict[event_nickname]) )
                    
            elif event_nickname in Event_Insertions:
                pd_event    = cls.mdl.parms.Event_dict[event_nickname]
                realiz_name = scenario_dict[event_nickname]
                realiz_id   = pd_event.loc[pd_event['value'] == realiz_name ].index.values[0]
                if event_nickname == 'v_3_del':
                    cls.id_vd_ins = realiz_id
                elif event_nickname == 'd_5_del':
                    cls.id_dj_ins = realiz_id
                else:
                    print( "Something vey bad happen with "+str(scenario_dict[event_nickname]) )
                
            elif event_nickname in Event_Dinucl:
                if event_nickname == 'vd_dinucl':
                    pd_event     = cls.mdl.parms.Event_dict[event_nickname]
                    str_sequence = scenario_dict[event_nickname]
                    list_id_seq  = list()
                    for str_nt in str_sequence:
                        realiz_name = str_nt
                        realiz_id   = pd_event.loc[pd_event['value'] == realiz_name ].index.values[0]
                        list_id_seq.append(realiz_id)
                    cls.vd_dinucl = list_id_seq
                
                elif event_nickname == 'dj_dinucl':
                    pd_event     = cls.mdl.parms.Event_dict[event_nickname]
                    str_sequence = scenario_dict[event_nickname]
                    list_id_seq  = list()
                    for str_nt in str_sequence:
                        realiz_name = str_nt
                        realiz_id   = pd_event.loc[pd_event['value'] == realiz_name ].index.values[0]
                        list_id_seq.append(realiz_id)
                    cls.dj_dinucl = list_id_seq
                
                else:
                    print("Something wrong with " + str(event_nickname) )

            elif event_nickname in ['mismatches']:
                if isinstance(scenario_dict[event_nickname], list):
                    cls.mismatches = scenario_dict[event_nickname]
                    cls.mismatcheslen = len(cls.mismatches)
                else:
                    print("mismatches aren't list")
            else:
                print( "Something bad happen!" )
        
        return cls
    
    
    def save_scenario_fasta(self, outfilename):
        ofileScen = open(outfilename, "w")
        ofileScen.write(self.str_scenario_fasta())
#        ofileScen.write("> "+str(self.seq_index)+", rank: "+str(self.scenario_rank)+ ", prob: "+str(self.scenario_proba_cond_seq)+"\n")
#        ofileScen.write( self.strSeq_index  + "\n" )
#        ofileScen.write( self.getV_fasta()  + "\n" )
#        ofileScen.write( self.getVD_fasta() + "\n" )
#        ofileScen.write( self.getD_fasta()  + "\n" )
#        ofileScen.write( self.getDJ_fasta() + "\n" )
#        ofileScen.write( self.getJ_fasta()  + "\n" )
        ofileScen.close()
        
    def str_scenario_fasta(self):
        strScenarioFasta = ""
        strScenarioFasta = strScenarioFasta + ">"+str(self.seq_index)+", rank: "+str(self.scenario_rank)+ ", prob: "+str(self.scenario_proba_cond_seq)+"\n"
        strScenarioFasta = strScenarioFasta + self.strSeq_index  + "\n" 
        strScenarioFasta = strScenarioFasta + self.getV_fasta()  + "\n" 
        strScenarioFasta = strScenarioFasta + self.getVD_fasta() + "\n" 
        strScenarioFasta = strScenarioFasta + self.getD_fasta()  + "\n"
        # FIXME: I THINK getDJ_fasta() is wrong (inverse order maybe)
        strScenarioFasta = strScenarioFasta + self.getDJ_fasta() + "\n" 
        strScenarioFasta = strScenarioFasta + self.getJ_fasta()  + "\n" 
        
        # FIXME: TEMPORAL 
        strScenarioFasta = strScenarioFasta + "> v_choice\n"
        strScenarioFasta = strScenarioFasta + self.getV_ntsequence() +"\n"
        
        strScenarioFasta = strScenarioFasta + "> d_gene\n"
        strScenarioFasta = strScenarioFasta + self.getD_ntsequence() +"\n"
        
        strScenarioFasta = strScenarioFasta + "> j_choice\n"
        strScenarioFasta = strScenarioFasta + self.getJ_ntsequence() +"\n"
        
        return strScenarioFasta

    
    #### V region methods
    def getV_fasta(self):
        strV_fasta = ""
        strV_fasta = strV_fasta + ">"+str(self.id_v_choice)+": "+ self.getV_gene_name()+ ", dels 3' = "+str(self.getV_3_dels()) + "\n"
        strV_fasta = strV_fasta + self.getV_Region()  + "\n"
        return strV_fasta
    
    def getV_gene_name(self):
        strEv  = 'v_choice'
        name_V = self.mdlParms.Event_dict[strEv].loc[self.id_v_choice]['name']
        return name_V
    
    def getV_ntsequence(self):
        strEv  = 'v_choice'
        seq_V  = self.mdlParms.Event_dict[strEv].loc[self.id_v_choice]['value']
        return seq_V
    
    def getV_3_dels(self):
        strEv  = 'v_3_del'
        n_v_3_del = self.mdlParms.Event_dict[strEv].loc[self.id_v_3_del]['value']
        return n_v_3_del
    
    def getV_Region(self):
        #seq_id=59
        strEv  = 'v_choice'
        seq_V  = self.mdlParms.Event_dict[strEv].loc[self.id_v_choice]['value']
        n_v_3_del = self.getV_3_dels()
        if n_v_3_del == 0 :
            return (seq_V)
        # FIXME: ADD palindromic insertions
        elif n_v_3_del < 0 :
            return (seq_V+'X'*n_v_3_del)
        else:
            return (seq_V[:-n_v_3_del])
   
    #### J region methods
    def getJ_fasta(self):
        strJ_fasta = ""
        strJ_fasta = strJ_fasta + ">"+str(self.id_j_choice)+": "+ self.getJ_gene_name() + ", dels 5' = "+str(self.getJ_5_dels()) + "\n"
        strJ_fasta = strJ_fasta + self.getJ_Region()  + "\n"
        return strJ_fasta
    
    def getJ_gene_name(self):
        strEv  = 'j_choice'
        name_J = self.mdlParms.Event_dict[strEv].loc[self.id_j_choice]['name']
        return name_J
    
    def getJ_ntsequence(self):
        strEv  = 'j_choice'
        seq_J  = self.mdlParms.Event_dict[strEv].loc[self.id_j_choice]['value']
        return seq_J
    
    def getJ_5_dels(self):
        strEv  = 'j_5_del'
        n_j_5_del = self.mdlParms.Event_dict[strEv].loc[self.id_j_5_del]['value']
        return n_j_5_del
    
    def getJ_Region(self):
        #seq_id=59
        strEv  = 'j_choice'
        seq_J  = self.mdlParms.Event_dict[strEv].loc[self.id_j_choice]['value']#.values            
        n_j_5_del = self.getJ_5_dels()
        if n_j_5_del == 0 :
            return (seq_J)
        # FIXME: ADD palindromic insertions
        elif n_j_5_del < 0 :
            return (seq_J+'X'*n_j_5_del)
        else:
            return (seq_J[n_j_5_del:])
    
    
    #### D region methods
    def getD_fasta(self):
        strD_fasta = ""
        strD_fasta = strD_fasta + ">"+str(self.id_d_gene)+": "+ self.getD_gene_name() \
                + ", "+str(self.id_d_5_del)+" dels 5' = "+str(self.getD_5_dels()) \
                + ", "+str(self.id_d_3_del)+" dels 3' = "+str(self.getD_3_dels()) + "\n"
        strD_fasta = strD_fasta + self.getD_Region()  + "\n"
        return strD_fasta
    
    def getD_gene_name(self):
        strEv  = 'd_gene'
        name_D = self.mdlParms.Event_dict[strEv].loc[self.id_d_gene]['name']
        return name_D
    
    def getD_ntsequence(self):
        strEv  = 'd_gene'
        seq_D  = self.mdlParms.Event_dict[strEv].loc[self.id_d_gene]['value']
        return seq_D
    
    def getD_5_dels(self):
        strEv  = 'd_5_del'
        n_d_5_del = self.mdlParms.Event_dict[strEv].loc[self.id_d_5_del]['value']
        return n_d_5_del
    
    def getD_3_dels(self):
        strEv  = 'd_3_del'
        n_d_3_del = self.mdlParms.Event_dict[strEv].loc[self.id_d_3_del]['value']
        return n_d_3_del
    
    def getD_Region(self):
        strEv  = 'd_gene'
        seq_D  = self.mdlParms.Event_dict[strEv].loc[self.id_d_gene]['value']#.values            
        n_d_5_del = self.getD_5_dels()
        n_d_3_del = self.getD_3_dels()
        if n_d_3_del == 0 :
            return (seq_D[n_d_5_del:])
        # FIXME: ADD palindromic insertions
        elif n_d_3_del < 0 :
            return (seq_D[n_d_5_del:]+'X'*n_d_3_del)
        else:
            return (seq_D[n_d_5_del:-n_d_3_del])
    
    
    #### VD region methods
    def getVD_fasta(self):
        strVD_fasta = ""
        strVD_fasta = strVD_fasta + ">"+str(self.id_vd_ins)+", VD insertions = "+str(self.getVD_ins()) + "\n"
        strVD_fasta = strVD_fasta + self.getVD_Region()  + "\n"
        return strVD_fasta
    
    def getVD_ins(self):
        strEv  = 'vd_ins'
        n_vd_ins = self.mdlParms.Event_dict[strEv].loc[self.id_vd_ins]['value']
        return n_vd_ins
    
    def getVD_Region(self):
        strEv        = 'vd_dinucl'
        seq_VD_dinucl  = self.mdlParms.Event_dict[strEv].loc[self.vd_dinucl]['value'].values   
        return (''.join(seq_VD_dinucl.tolist())) 
    
    #### DJ region methods
    def getDJ_fasta(self):
        strDJ_fasta = ""
        strDJ_fasta = strDJ_fasta + ">"+str(self.id_dj_ins)+", DJ insertions = "+str(self.getDJ_ins()) + "\n"
        strDJ_fasta = strDJ_fasta + self.getDJ_Region()  + "\n"
        return strDJ_fasta
    
    def getDJ_ins(self):
        strEv  = 'dj_ins'
        n_dj_ins = self.mdlParms.Event_dict[strEv].loc[self.id_dj_ins]['value']
        return n_dj_ins
    
    def getDJ_Region(self):
        strEv        = 'dj_dinucl'
        seq_DJ_dinucl  = self.mdlParms.Event_dict[strEv].loc[self.dj_dinucl]['value'].values   
        return (''.join(seq_DJ_dinucl.tolist())) 
    
    # FIXME: CHANGE NAME TO get_ScenarioProb.
    def get_EventProb(self):
        ###### v_choice
        strEvent = 'v_choice'
        da_event = self.mdl.xdata[strEvent]
        p_V = da_event[{strEvent: self.id_v_choice}]
        #print("p_V = ", p_V)
        #p_V = da_event.where(da_event['lbl__'+strEvent] == bs.getV_gene_name() ,drop=True)
        
        ###### j_choice
        strEvent = 'j_choice'
        da_event = self.mdl.xdata[strEvent]
        p_J = da_event[{strEvent: self.id_j_choice}]
        #print("p_J = ", p_J)
        #p_J = da_event.where(da_event['lbl__'+strEvent] == bs.getJ_gene_name() ,drop=True)
        
        ###### d_gene
        strEvent = 'd_gene'
        da_event = self.mdl.xdata[strEvent]
        p_DgJ = da_event[{'d_gene':self.id_d_gene, 'j_choice':self.id_j_choice}]
        #print("p_DgJ = ", p_DgJ)
        #p_DgJ
        
        
        ###### v_3_del
        strEvent = 'v_3_del'
        da_event = self.mdl.xdata[strEvent]
        p_V_3_del = da_event[{'v_3_del':self.id_v_3_del, 'v_choice':self.id_v_choice}]
        #print("p_V_3_del = ", p_V_3_del)
        #p_V_3_del
        
        
        ###### j_5_del
        strEvent = 'j_5_del'
        da_event = self.mdl.xdata[strEvent]
        p_J_5_del = da_event[{'j_5_del':self.id_j_5_del, 'j_choice':self.id_j_choice}]
        #print("p_J_5_del = ", p_J_5_del)
        #p_J_5_del
        
        ###### d_5_del
        strEvent = 'd_5_del'
        da_event = self.mdl.xdata[strEvent]
        p_D_5_del = da_event[{'d_5_del':self.id_d_5_del, 'd_gene':self.id_d_gene}]
        #print("p_D_5_del = ", p_D_5_del)
        #p_D_5_del
        
        ###### d_3_del
        strEvent = 'd_3_del'
        da_event = self.mdl.xdata[strEvent]
        p_D_3_del = da_event[{'d_3_del':self.id_d_3_del, 'd_5_del':self.id_d_5_del, 'd_gene':self.id_d_gene}]
        #print("p_D_3_del = ", p_D_3_del)
        #p_D_3_del
        
        ###### vd_ins
        strEvent = 'vd_ins'
        da_event = self.mdl.xdata[strEvent]
        p_VD_ins = da_event[{'vd_ins':self.id_vd_ins}]
        #print("p_VD_ins = ", p_VD_ins)
        
        
        ###### vd_dinucl
        strEvent = 'vd_dinucl'
        da_event = self.mdl.xdata[strEvent]
        # Get the last nucleotide of V region (after deletions)
        
        str_prev_nt = self.getV_Region()[-1]
        pd_tmp = self.mdl.parms.Event_dict[strEvent]
        prev_nt = pd_tmp.loc[pd_tmp['value'] == str_prev_nt].index.values[0]
        
        # for each nucleotide on inserted list
        Num_nt = 4 # 4 nucleotides A, C, G, T 
        p_VD_dinucl = 1
        for curr_nt in self.vd_dinucl:
            id_dinucl   = prev_nt*Num_nt + curr_nt
            prob_tmp    = da_event[{'vd_dinucl':id_dinucl}]
            p_VD_dinucl = p_VD_dinucl * prob_tmp
            #print(prev_nt, curr_nt, id_dinucl, prob_tmp, p_VD_dinucl)
            prev_nt = curr_nt
        
        #
        #print("p_VD_dinucl = ", p_VD_dinucl)
        
        ###### dj_ins
        strEvent = 'dj_ins'
        da_event = self.mdl.xdata[strEvent]
        p_DJ_ins = da_event[{'dj_ins':self.id_dj_ins}]
        #print("p_DJ_ins = ", p_DJ_ins)
        
        ###### dj_dinucl
        strEvent = 'dj_dinucl'
        da_event = self.mdl.xdata[strEvent]
        # Get the last nucleotide of V region (after deletions)
        
#        str_prev_nt = (self.getV_Region() + self.getVD_Region() + self.getD_Region() )[-1]
        str_prev_nt = (self.getJ_Region() )[0]
        pd_tmp  = self.mdl.parms.Event_dict[strEvent]
        prev_nt = pd_tmp.loc[pd_tmp['value'] == str_prev_nt].index.values[0]
        #print("prev_nt : ", prev_nt)
        
        # for each nucleotide on inserted list
        Num_nt = 4 # 4 nucleotides A, C, G, T 
        p_DJ_dinucl = 1
        #self.dj_dinucl = self.dj_dinucl[::-1]
        for curr_nt in self.dj_dinucl:
            id_dinucl   = prev_nt*Num_nt + curr_nt
            prob_tmp    = da_event[{'dj_dinucl':id_dinucl}]
            p_DJ_dinucl = p_DJ_dinucl * prob_tmp
            #print(prev_nt, curr_nt, id_dinucl, prob_tmp, p_DJ_dinucl)
            prev_nt = curr_nt
        
        #print("p_DJ_dinucl = ", p_DJ_dinucl)
        
        
        p_vecE = p_V*p_J*p_DgJ*p_V_3_del*p_J_5_del*p_D_5_del*p_D_3_del*\
                p_VD_ins*p_VD_dinucl*p_DJ_ins*p_DJ_dinucl
        
        return p_vecE.values
    
    def get_DictNicknameProbs(self):
        dictNicknameProbs       =  dict()
        {
            "v_choice"      : self.id_v_choice   , \
            "j_choice"      : self.id_j_choice   , \
            "d_gene"        : self.id_d_gene     , \
            "v_3_del"       : self.id_v_3_del    , \
            "d_5_del"       : self.id_d_5_del    , \
            "d_3_del"       : self.id_d_3_del    , \
            "j_5_del"       : self.id_j_5_del    , \
            "vd_ins"        : self.id_vd_ins     , \
            "vd_dinucl"     : self.vd_dinucl     , \
            "dj_ins"        : self.id_dj_ins     , \
            "dj_dinucl"     : self.dj_dinucl     , \
            "mismatches"    : self.mismatches    , \
            "mismatcheslen" : self.mismatcheslen
        }
        ###### v_choice
        strEvent = 'v_choice'
        da_event = self.mdl.xdata[strEvent]
        p_V = da_event[{strEvent: self.id_v_choice}]
        dictNicknameProbs[strEvent] = p_V 
        #print("p_V = ", p_V)
        #p_V = da_event.where(da_event['lbl__'+strEvent] == bs.getV_gene_name() ,drop=True)
        
        ###### j_choice
        strEvent = 'j_choice'
        da_event = self.mdl.xdata[strEvent]
        p_J = da_event[{strEvent: self.id_j_choice}]
        dictNicknameProbs[strEvent] = p_J 
        #print("p_J = ", p_J)
        #p_J = da_event.where(da_event['lbl__'+strEvent] == bs.getJ_gene_name() ,drop=True)
        
        ###### d_gene
        strEvent = 'd_gene'
        da_event = self.mdl.xdata[strEvent]
        p_DgJ = da_event[{'d_gene':self.id_d_gene, 'j_choice':self.id_j_choice}]
        dictNicknameProbs[strEvent] = p_DgJ
        #print("p_DgJ = ", p_DgJ)
        #p_DgJ
        
        
        ###### v_3_del
        strEvent = 'v_3_del'
        da_event = self.mdl.xdata[strEvent]
        p_V_3_del = da_event[{'v_3_del':self.id_v_3_del, 'v_choice':self.id_v_choice}]
        dictNicknameProbs[strEvent] = p_V_3_del
        #print("p_V_3_del = ", p_V_3_del)
        #p_V_3_del
        
        
        ###### j_5_del
        strEvent = 'j_5_del'
        da_event = self.mdl.xdata[strEvent]
        p_J_5_del = da_event[{'j_5_del':self.id_j_5_del, 'j_choice':self.id_j_choice}]
        dictNicknameProbs[strEvent] = p_J_5_del
        
        ###### d_5_del
        strEvent = 'd_5_del'
        da_event = self.mdl.xdata[strEvent]
        p_D_5_del = da_event[{'d_5_del':self.id_d_5_del, 'd_gene':self.id_d_gene}]
        dictNicknameProbs[strEvent] = p_D_5_del
        
        ###### d_3_del
        strEvent = 'd_3_del'
        da_event = self.mdl.xdata[strEvent]
        p_D_3_del = da_event[{'d_3_del':self.id_d_3_del, 'd_5_del':self.id_d_5_del, 'd_gene':self.id_d_gene}]
        dictNicknameProbs[strEvent] = p_D_3_del
        
        ###### vd_ins
        strEvent = 'vd_ins'
        da_event = self.mdl.xdata[strEvent]
        p_VD_ins = da_event[{'vd_ins':self.id_vd_ins}]
        dictNicknameProbs[strEvent] = p_VD_ins
        
        
        ###### vd_dinucl
        strEvent = 'vd_dinucl'
        da_event = self.mdl.xdata[strEvent]
        # Get the last nucleotide of V region (after deletions)
        
        str_prev_nt = self.getV_Region()[-1]
        pd_tmp = self.mdl.parms.Event_dict[strEvent]
        prev_nt = pd_tmp.loc[pd_tmp['value'] == str_prev_nt].index.values[0]
        
        # for each nucleotide on inserted list
        Num_nt = 4 # 4 nucleotides A, C, G, T 
        p_VD_dinucl = 1
        for curr_nt in self.vd_dinucl:
            id_dinucl   = prev_nt*Num_nt + curr_nt
            prob_tmp    = da_event[{'vd_dinucl':id_dinucl}]
            p_VD_dinucl = p_VD_dinucl * prob_tmp
            #print(prev_nt, curr_nt, id_dinucl, prob_tmp, p_VD_dinucl)
            prev_nt = curr_nt
        
        dictNicknameProbs[strEvent] = p_VD_dinucl
        
        ###### dj_ins
        strEvent = 'dj_ins'
        da_event = self.mdl.xdata[strEvent]
        p_DJ_ins = da_event[{'dj_ins':self.id_dj_ins}]
        dictNicknameProbs[strEvent] = p_DJ_ins
        
        ###### dj_dinucl
        strEvent = 'dj_dinucl'
        da_event = self.mdl.xdata[strEvent]
        # Get the last nucleotide of V region (after deletions)
        
        str_prev_nt = (self.getV_Region() + self.getVD_Region() + self.getD_Region() )[-1]
        pd_tmp  = self.mdl.parms.Event_dict[strEvent]
        prev_nt = pd_tmp.loc[pd_tmp['value'] == str_prev_nt].index.values[0]
        #print("prev_nt : ", prev_nt)
        
        # for each nucleotide on inserted list
        Num_nt = 4 # 4 nucleotides A, C, G, T 
        p_DJ_dinucl = 1
        for curr_nt in self.dj_dinucl:
            id_dinucl   = prev_nt*Num_nt + curr_nt
            prob_tmp    = da_event[{'dj_dinucl':id_dinucl}]
            p_DJ_dinucl = p_DJ_dinucl * prob_tmp
            #print(prev_nt, curr_nt, id_dinucl, prob_tmp, p_DJ_dinucl)
            prev_nt = curr_nt
        
        dictNicknameProbs[strEvent] = p_DJ_dinucl
        
        return dictNicknameProbs

    def get_ErrorProb(self):
        r = float(self.mdl.parms.ErrorRate['SingleErrorRate'])
        L = len(self.strSeq_index)
        print("error rate: ", r, "n mismatches", self.mismatcheslen)
        #return r**(self.mismatcheslen)
        return (r/3)**(self.mismatcheslen) # * (1-r)**( L - self.mismatcheslen)
    
